/*

    Copyright (c) 1989, 1990 University of Southern California

	Paul Raveling
	USC Information Sciences Institute
	4676 Admiralty Way
	Marina del Rey, California 90292


	The authors grant without fee permission to use, copy, modify,
	distribute, and sell this software for any purpose, provided:

	--  The above copyright notice appears in all copies.

	--  This disclaimer appears in all source code copies.

	--  The names of ISI and USC are not used in advertising or
	    publicity  pertaining  to distribution  of the software
	    without prior specific written permission from ISI.

	ISI provides this software freely as a public service.  It is
	NOT a commercial product,  and therefore is not subject to an
	an implied  warranty  of  merchantability  or  fitness  for a
	particular purpose.  ISI provides it as is, without warranty.

	This software  is furnished  only on the basis that any party
	who receives it indemnifies  and holds harmless  the original
	authors and their sponsors  against any claims,  demands,  or
	liabilities  connected  with  using  it,   furnishing  it  to
	others, or providing it to a third party.


	The author  of this software is  Paul Raveling.   Development
	used the author's after-work hours and ISI's facilities.

	Thanks are due to ISI for allowing  such extracurricular work
	and  to Hewlett Packard,   whose  University  Grants  program
	provided  the 9000/370  workstation  and  supporting software
	which was the platform for this work.
*/

/*0
===============================================================================
=======                                                                 =======
=======		   Module Imgcolor  --  Manage Colors			=======
=======			      File imgcolor.c				=======
=======                                                                 =======
===============================================================================


  Functions:

	Manage image colors.

	This includes commonly used functions related for reducing the
	chromatic content of an image.  It also includes less commonly
	used functions for supporting multiple color maps with a
	single set of pixel data.

	Beware:  code for color dithering is in an intermediate
	stage between something that worked and something that
	was fast.  It's not currently used but some variant of it
	will probably revive in the future.


  Public Routines:

	init_colman	Initialize image for color management
	dither_rgb	Dither an RGB image


  Local Routines:

	expand_color	Expand a color definition
	nearest_color	Find nearest mapped color


  Module History:

	6-15-90		Paul Raveling	V1.3
	Deleted old experimental code for color dithering; dithering
	is now incorporated in quantization (see quantize.c).

	3-7-90		Paul Raveling
	Finalized for version 1.2 by updating copyright notice.
	Quantization is in module quantize, not here.

	12-12-89	Paul Raveling
	Beginning mods for new color quantization algorithm.  Individual
	changes won't be logged until the prototype is up.

	11-28-89	Paul Raveling
	Updated copyright/legal disclaimer notice to suit
	X Consortium requirements

	11-22-89	Paul Raveling
	Finalized version 1.1 for contribution to X11R4.
	Changes not documented in this header have been made
	within the last 3 weeks.

	8-21-89		Paul Raveling
	Fixed bug in init_colman:  One reference was to the wrong
	pixel pointer when scanning an RGB image.

	3-24-89		Paul Raveling
	Created module.
*/



#include <stdio.h>
#include <imglib/imglib.h>

#define	reg	 register


extern	char	*malloc ();		/*  Allocate memory		*/
extern	char	*calloc ();		/*  Allocate memory differently	*/


/*1
===============================================================================
=======                                                                 =======
=======	    init_colman  --  Initialize Image for Color Management	=======
=======                                                                 =======
===============================================================================

  Invocation:

	status = init_colman ( image )


  Functions:

	Initialize color management data for a color-mapped image.
	The image type can be either mapped_image or rgb_image,
	but an RGB image must have a color map, which would otherwise
	be optional.  RGB pixel values must have color map indices
	in their tag fields.

	This routine must be called after the image is created
	and before most other color management routines are invoked
	on it.

	Specific data initialized are:

	--  A color info queue entry for the current color map.
	    If queue entries exist, init_colman updates the one
	    at the head of the queue.  If none exist, it creates
	    a new queue entry.

	--  Pixel frequency statistics in the current color map
	    are initialized by scanning the pixel data.  While
	    doing this, pixels are checked for invalid color map
	    indices.  If any exist, a summary message is sent to
	    stderr and the faulty pixel values are replaced by 0.


  Input:

	image		Img	*
			Pointer to Img structure for image to delete


  Output:

	status		int
	<return value>	0 if initialization succeeded;
			1 if init completed but invalid pixels were found
			2 if init failed due to insufficient memory
			  or incorrect image format
*/



int	 init_colman ( image )

	Img	*image;		/*  Ptr to image to initialize	*/

{
	int	 bad_pixels;	/*  Count of invalid pixels		*/
	Mapinfo	*ci;		/*  Ptr to color info queue entry	*/
	Byte	 pixel;		/*  Current pixel value			*/
	Byte	*pp;		/*  Pixel pointer for mapped image	*/
	Rgbpix	*rgbp;		/*  Pixel pointer for RGB image		*/
	int	 i;		/*  Universal index			*/




/*::::::::::::::::::::::::::::::::::::::::*/
/*	Macro to abort with message	  */
/*::::::::::::::::::::::::::::::::::::::::*/

#define	abort_init(msg) {fprintf (stderr, \
 "init_colman:  %s, can't initialize color management\n", msg); return 2;}



/*-----------------------------------------------------------------*/
/*	Verify that image has both a color map and pixel data	   */
/*-----------------------------------------------------------------*/

if ( (image->imgtype != mapped_image) && (image->imgtype != rgb_image) )
   abort_init("Wrong image type")

if ( (image->colormap == NULL) || (image->num_colors == 0) )
   abort_init("No color map")

if ( (image->pixdata == NULL) || (image->num_pixels == 0) )
   abort_init("No pixel data")



/*----------------------------------------------*/
/*	Initialize Color Info Queue Entry	*/
/*----------------------------------------------*/

ci = image -> cmiq;
if ( ci == NULL )		/*  If no quene entry exists yet	*/
   {				/*     Allocate and enqueue one		*/
   ci = (Mapinfo *) malloc (sizeof(Mapinfo));
   if (ci == NULL)
      abort_init("Insufficient memory to create color info structure")

   image -> cmiq = ci;		/*  Define entire color info queue	*/
   ci -> next_map = NULL;	/*  to be the new entry			*/
   }

ci -> map	= image -> colormap;	/*  Set pointer to color map	*/
ci -> mapcolors = image -> num_colors;	/*  Set count of colors in map	*/
ci -> mapunique = 0;			/*  Preset count of uniques = 0	*/




/*-----------------------------------------------*/
/*	   Check pixels for validity and	 */
/*	   Generate color frequency stats	 */
/*-----------------------------------------------*/

bad_pixels = 0;				/*  Preset for no bad pixels	*/

for ( i=0; i < image->num_colors; ++i )	/*  Preset color frequencies	*/
    image->colormap[i].freq = 0;	/*	to 0			*/

if ( image->imgtype == mapped_image )	/*  Count pixels for each color	*/
   {
   pp = image->pixdata;			/*  ****  Mapped Image  ****	*/
   i  = image->num_pixels;
   while ( i-- > 0 )			/*  For each pixel...		*/
	 {
	 pixel = *pp++;
	 if ( pixel >= image->num_colors )  /*  If pixel is invalid	*/
	    {
	    ++ bad_pixels;		/*	  count it and		*/
	    *(pp-1) = 0;		/*	  replace it with 0	*/
	    pixel   = 0;
	    }
	 ++ image->colormap[pixel].freq;/*  Increment color count for	*/
	 }				/*  this pixel's color		*/
   }
else
   {
   rgbp = (Rgbpix *) image->pixdata;	/*  ****  RGB image  ****	*/
   i  = image->num_pixels;
   while ( i-- > 0 )			/*  For each pixel...		*/
	 {
	 pixel = (*rgbp++ & pixel_tag) >> pixel_tag_shift;
	 if (pixel >= image->num_colors)   /*  If pixel is invalid	*/
	    {
	    ++ bad_pixels;		/*	  count it and		*/
	    pixel   = 0;		/*	  replace it with 0	*/
	    *(rgbp-1) &= ~pixel_tag;
	    }
	 ++ image->colormap[pixel].freq;/*  Increment color count for	*/
	 }				/*  this pixel's color		*/
   }

image->state |= cm_inited;		/*  Set state flag to show	*/
					/*  "color mgmnt initialized"	*/

if ( bad_pixels != 0 )			/*  If any pixels were invalid	*/
   {					/*     report how many were bad	*/
   fprintf ( stderr, "init_colman:  %d invalid pixels replaced by 0\n" );
   return 1;
   }

return 0;
}

